/*
 * debug.c
 *
 * A dynamic debugging framework for Kaffe.  Through the magic
 * of hideous macros, we can control the debugging output of
 * Kaffe in a couple of ways.  First, it can be completely
 * disabled in which case all of the code "just goes away",
 * or, if its enabled, the area of code to debug is dynamically
 * chosen at run time.
 *
 * Copyright (c) 1997 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by Patrick Tullmann <tullmann@cs.utah.edu>, 1997
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "debug.h"

/* Default debugging mask to use (if debug is enabled) */
#define DEFAULT_DEBUG_MASK	DBG_NONE

#if defined(NDEBUG) || !defined(DEBUG)
/* Don't waste space with the debugging functions */

void dbgSetMask(int m) { }
void dbgSetMaskStr(char *s) { }

#else /* Actually define the functions */

int kaffevmDebugMask = DEFAULT_DEBUG_MASK;

void dbgSetMask(int mask)
{
	kaffevmDebugMask = mask;
}

/*
 * Associate strings with an option or set of
 * options.
 */
static struct debug_opts {
      char *name;
      int mask;
} debug_opts[] = {
#define D(name) { #name, DBG_ ## name }
	D(NONE),
	D(GC),
	D(SETUP),
	D(CLASS_LOAD),
	D(CLASS_LINK),
	D(CLASS_PREPARE),
	D(CLASS),
	D(ZIPFILE),
	D(NATIVE_METH),
	D(LOOKUP_METH),
	D(LOOKUP_FIELD),
	D(LOOKUP_EXC),
	D(LOOKUP),
	D(INSTR),
	D(METHOD_CALL),
	D(TRANSLATE),
	D(CLASS_STATE),
	D(ALLOC),
	D(CHECKS),
	D(THREADOP),
	D(EXCEPTION),
	D(NOCATCHSIGNAL),
	D(NPM_SPACE),
	D(NPM_SETUP),
	D(LOCKS),
	D(NOIDLE),
	D(NATIVE_IO),
	D(NATIVE_LANG_THREAD),
	D(CLASS_READ),
	D(ALL),
	{ "trace", DBG_INSTR|DBG_CHECKS|DBG_CLASS_LOAD|DBG_METHOD_CALL },
	{ "csw", DBG_THREADOP|DBG_LOCKS },
	{ "init", DBG_SETUP|DBG_CLASS|DBG_ZIPFILE }
};
	

void dbgSetMaskStr(char *mask_str)
{
	int i;
	char *separators = "|,";
	char *opt;
	
	opt = strtok(mask_str, separators);

	if (!opt) {
		kaffevmDebugMask = DEFAULT_DEBUG_MASK;
		return;
	}

	/* Special target 'list' lists all the defined options */
	if (!strcmp(opt, "list")) {
		printf("Available debug opts: \n");
		printf("  %-15s\t%8s\n", "Option", "Mask");
		for (i = 0; i < (sizeof debug_opts)/(sizeof(debug_opts[0])); i++)
			printf("  %-15s\t%8X\n", debug_opts[i].name, debug_opts[i].mask);
		printf("Exiting.\n");
		exit(0);
	}
	

	while (opt) {
		for (i = 0; i < (sizeof debug_opts)/(sizeof(debug_opts[0])); i++)
			/* probably should strcasecmp() or stricmp() */
			if (!strcmp(opt, debug_opts[i].name)) {
				kaffevmDebugMask |= debug_opts[i].mask;
				break;
			}
		
		/* Be polite. */
		if (i == (sizeof debug_opts)/(sizeof(debug_opts[0])))
			fprintf(stderr, "Unknown flag (%s) passed to -vmdebug\n",
				opt);

		/* Get next opt */
		opt = strtok(NULL, separators);
	}
}
#endif

