/*
 * debug.h
 *
 * A dynamic debugging framework for Kaffe.  Through the magic
 * of hideous macros, we can control the debugging output of
 * Kaffe in a couple of ways.  First, it can be completely
 * disabled in which case all of the code "just goes away",
 * or, if its enabled, the area of code to debug is dynamically
 * chosen at run time.
 *
 * Copyright (c) 1997 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by Patrick Tullmann <tullmann@cs.utah.edu>, 1997
 */
#ifndef __kaffevm_debug_h
#define __kaffevm_debug_h

/* Defines what debugging output is seen. Needs to be 32-bit. */
extern int kaffevmDebugMask;

#if defined(NDEBUG) || !defined(DEBUG)
/* --- Debugging is NOT enabled --- */

/* Replace the mask set functions with empty macros */
# define DBG(mask, statement)  { ((void)0); }

# define DBGIFNOT(mask, statement)  { ((void)0); }

# define DBGEXPR(mask, expr, default) (default)

# define DBGPRINT(mask, printfargs) { ((void)0); } 

#else 
/* --- Debugging is enabled --- */

/* Set the debugging mask to use. (give the mask) */
void dbgSetMask(int mask);

/* Set the debugging mask to use. (give a string, useful for
 * parsing a command line option. */
void dbgSetMaskStr(char *mask_str);

/* Debug Masks: (1 bit per) */
/* NOTE: Many of these names are silly, and there is probably a
 * better organization.  Feel free to throw away names and fixup
 * the groupings. */
# define DBG_NONE		(0)
# define DBG_SETUP		(1<<1)
# define DBG_CLASS_LOAD		(1<<2)
# define DBG_CLASS_LINK		(1<<3)
# define DBG_CLASS_PREPARE	(1<<4)
# define DBG_CLASS_STATE        (1<<24)
# define DBG_CLASS_ERR		(1<<12)
# define DBG_CLASS_READ         (1<<29)
# define DBG_CLASS 		(DBG_CLASS_LOAD|DBG_CLASS_LINK|DBG_CLASS_PREPARE|DBG_CLASS_STATE|DBG_CLASS_ERR)
# define DBG_ZIPFILE		(1<<5)
# define DBG_NATIVE_METH	(1<<6) 
  /* lookup.c: */
# define DBG_LOOKUP_METH	(1<<7)
# define DBG_LOOKUP_FIELD       (1<<8)
# define DBG_LOOKUP_EXC         (1<<9)
# define DBG_LOOKUP             (DBG_LOOKUP_METH|DBG_LOOKUP_FIELD|DBG_LOOKUP_EXC)
# define DBG_METHOD		(1<<10)
# define DBG_FIELD              (1<<11) 
                                // 12 is above
# define DBG_INSTR		(1<<13)
# define DBG_METHOD_CALL	(1<<14)
# define DBG_TRANSLATE		(1<<15)
# define DBG_ALLOC		(1<<16)
# define DBG_GC			(1<<17)
# define DBG_CHECKS		(1<<18) /* see intrp/checks.h --- stack validity checks */
# define DBG_THREADOP		(1<<19)
# define DBG_EXCEPTION		(1<<20)
# define DBG_NOCATCHSIGNAL      (1<<21)
# define DBG_NPM_SPACE          (1<<22)
# define DBG_NPM_SETUP		(1<<23)
                                // 24 is above
# define DBG_LOCKS              (1<<25)
# define DBG_NOIDLE             (1<<26) // don't let "idle thread" run
# define DBG_NATIVE_IO		(1<<27) 
# define DBG_NATIVE_LANG_THREAD	(1<<28) 
				// 29 is above
# define DBG_PROFILING          (1<<30)
# define DBG_ALL		(0xFFFFFFFF)

/* Debug macros that are selected with the above flags. */

# define DBG(mask, statement) {				\
	if ((mask)&(kaffevmDebugMask)) {		\
		statement;				\
	}}

# define DBGIFNOT(mask, statement) {			\
	if (!((mask)&(kaffevmDebugMask))) {		\
		statement;				\
	}}

# define DBGEXPR(mask, expr, default)			\
	(((mask)&(kaffevmDebugMask))&(expr))

/* If only all the world used GNU CPP, which lets macros take a variable
 * number of arguments... */
# define DBGPRINT(mask, printfargs) {			\
		if ((mask)&kaffevmDebugMask) {		\
			printf printfargs;		\
		}}

		
#endif /* defined(NDEBUG) || !defined(DEBUG) */

#endif /* __kaffevm_debug_h */
