/*
 * machine.c
 * Java virtual machine interpreter.
 *
 * Copyright (c) 1996 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>
 */

#include "config.h"
#include "config-std.h"
#include "config-math.h"
#include "config-mem.h"
#include <setjmp.h>
#include "gtypes.h"
#include "bytecode.h"
#include "slots.h"
#include "icode.h"
#include "access.h"
#include "object.h"
#include "constants.h"
#include "gc.h"
#include "machine.h"
#include "classMethod.h"
#include "lookup.h"
#include "soft.h"
#include "exception.h"
#include "external.h"
#include "baseClasses.h"
#include "thread.h"
#include "checks.h"
#include "errors.h"
#include "md.h"
#include "code-analyse.h"
#include "debug.h"

/*
 * Define information about this engine.
 */
char* engine_name = "Interpreter";
char* engine_version = KVER;

extern const uint8 insnLen[];

#define	define_insn(code)	break;					\
				case code:				\
				DBG(DBG_INSTR, printf("%03d: %s\n", pc, #code); )

/* Define CREATE_NULLPOINTER_CHECKS in md.h when your machine cannot use the
 * MMU for detecting null pointer accesses */
#if defined(CREATE_NULLPOINTER_CHECKS)
#define CHECK_NULL(_i, _s, _n)                                  \
      cbranch_ref_const_ne((_s), 0, reference_label(_i, _n)); \
      softcall_nullpointer();                                 \
      set_label(_i, _n)
#else
#define CHECK_NULL(_i, _s, _n)
#endif

/* Misc stuff */
Hjava_lang_Object* exceptionObject;

#ifdef DEBUG
#define MAXDEPTHPRINT 60
static char *nSpaces(int n) {
#define BLANKS_STRING_LEN 120
	static char blanks[BLANKS_STRING_LEN+1];
	assert(n < BLANKS_STRING_LEN);
	if (n < 0) return blanks;
	(n > 0) ? (blanks[n-1] = ' ') : (blanks[0] = '\0');
	(n < BLANKS_STRING_LEN) ? (blanks[n+1] = ' ') : (blanks[BLANKS_STRING_LEN] = '\0');

	blanks[n] = '\0';
	return blanks;
}

static int callDepth = 0;

void dbgMethodCallDone()
{
	callDepth--;
	if (callDepth < MAXDEPTHPRINT)
	{
		printf("%s%d-",
		       nSpaces(callDepth), callDepth % 10);
	}
}

const char *dbgMethodReturnVal(const Utf8Const *signature, const slots *retval)
{
	static char str[100];
	char *sig = signature->data;
	
	assert(signature);
	assert(retval);

	// XXX won't work with real Unicode names
	while(*sig && ((*sig) != ')'))
		sig++;

	assert(*sig == ')'); // wacky signature has no ')' ??? 
	sig++;
	
	// It should now point to the return type... 
	switch(*sig)
	{
	case 'V': // Void
		str[0] = '\0';
		break;
	case 'B': // Byte
	case 'I': // Integer
	case 'S': // Short
		sprintf(str, "%d (==0x%x)",
			retval->v.tint, retval->v.tint);
		break;
	case 'C': // XXX Java Char
		sprintf(str, "%d (==0x%x, ==%c)",
			retval->v.tint, retval->v.tint,
			(char)(retval->v.tint));
		break;
	case 'Z': // Boolean
		sprintf(str, "%s",
			retval->v.tint ? "true" : "false"); // XXX ??
		break;
	case 'J': // Long
		sprintf(str, "%ld (==0x%lx)", (signed long)(retval->v.tlong),
			(signed long)(retval->v.tlong));
		break;
	case 'F': // Float
		sprintf(str, "%g",
			retval->v.tfloat);
		break;
	case 'D': // Double
		sprintf(str, "%g",
			retval->v.tdouble);
		break;
	case 'L': // Object reference
	case '[': // Array ref
		sprintf(str, "%p",
			retval->v.taddr);
		break;
	default:
		sprintf(str, "ERROR ret type: %c?", *sig);
	}

	return str;
}
#endif


void
virtualMachine(methods* meth, slots* volatile arg, slots* retval)
{
	Hjava_lang_Object* volatile mobj;
	vmException mjbuf;
	accessFlags methaccflags;
	char* str;

	/* If these can be kept in registers then things will go much
	 * better.
	 */
	register bytecode* code;     /* raw byte codes for the method */
	register slots* lcl;         /* stack space and locals for the method */
	register slots* sp;          /* stack pointer.  Initialzed to end of lcl block */
	register uintp pc;           /* program counter.  index in code of current isn. */
	volatile register uintp npc; /* next pc.  index in code of next isn. */

	/* Misc machine variables */
	/* jint cc; */
	jlong lcc;
	jlong tmpl;
	slots tmp[1];
	slots tmp2[1];
	slots mtable[1];

	/* Variables used directly in the machine */
	int wide;
	int32 idx;
	jint low;
	jint high;

	/* Call, field and creation information */
	callInfo cinfo;
	Field* field;
	Hjava_lang_Class* crinfo;
	/* exceptionInfo info; */


	/* If this is native, then call the real function */
	methaccflags = meth->accflags;
	if (methaccflags & ACC_NATIVE) {
		if (meth->ncode == 0) {
			native(meth);
		}

		DBG(DBG_NATIVE_METH|DBG_METHOD_CALL,
		    if (callDepth < MAXDEPTHPRINT)
		    {
			    printf("%s%d+native %s.%s%s.\n",
				   nSpaces(callDepth), callDepth % 10,
				   meth->class->name->data,
				   meth->name->data, meth->signature->data);
		    }
		    callDepth++;
		    fflush(stdout);
		    );

		(*(void(*)(void* volatile, void*))meth->ncode)(arg, retval);

		DBG(DBG_NATIVE_METH|DBG_METHOD_CALL,
		    dbgMethodCallDone();			
    
		    if (callDepth < MAXDEPTHPRINT)
		    {
			    printf("native %s.%s%s:%s\n",
				   meth->class->name->data,
				   meth->name->data, meth->signature->data,
				   dbgMethodReturnVal(meth->signature, retval));
		    }
		    
		    );
		return;
	}

	/* Verify method if required */
	if ((methaccflags & ACC_VERIFIED) == 0) {
		verifyMethod(meth);
		tidyVerifyMethod();
	}

	/* Allocate stack space and locals. */
	lcl= alloca(sizeof(slots) * (meth->localsz + meth->stacksz));

	mobj = 0;
	pc = 0;
	npc = 0;

	/* If we have any exception handlers we must prepare to catch them.
	 * We also need to catch if we are synchronised (so we can release it).
	 */
	mjbuf.pc = 0;
	mjbuf.mobj = (Hjava_lang_Object* /*volatile*/)mobj;
	mjbuf.meth = meth;
	if (currentThread != NULL && unhand(currentThread)->PrivateInfo != 0) {
		mjbuf.prev = TCTX(currentThread)->exceptPtr;
		TCTX(currentThread)->exceptPtr = &mjbuf;
	}
 	if (meth->exception_table != 0 || (methaccflags & ACC_SYNCHRONISED)) {
		if (setjmp(mjbuf.jbuf) != 0) 
		{
			assert(TCTX(currentThread)->exceptPtr == &mjbuf);
			npc = mjbuf.pc;
			sp = &lcl[meth->localsz + meth->stacksz - 1];
			sp->v.taddr = (void*)exceptionObject;
			goto restart; /*forward*/
		}
	}

	/* Calculate number of arguments */
	str = meth->signature->data;
	idx = sizeofSig(&str, false) + (methaccflags & ACC_STATIC ? 0 : 1);

	DBG(DBG_METHOD_CALL,
	    if (callDepth < MAXDEPTHPRINT)
	    {
		    printf("%s%d+method %s.%s%s.\n",
			   nSpaces(callDepth), callDepth % 10,
			   meth->class->name->data, meth->name->data,
			   meth->signature->data);
	    }
	    callDepth++;
	    fflush(stdout);
	    );
	
	/* Copy args from arg parameter into the local stack space (lcl) */
	sp = lcl;
	arg = &arg[idx-1];
	for (low = 0; low < idx; low++) {
		*(sp++) = *(arg--);
	}

	/* Sync. if required */
	if (methaccflags & ACC_SYNCHRONISED) { 
		if (methaccflags & ACC_STATIC) {
			mobj = &meth->class->head;
		}
		else {
			mobj = (Hjava_lang_Object*)lcl[0].v.taddr;
		}
		soft_monitorenter((Hjava_lang_Object* /*volatile*/)mobj);
	}       

	/* Init stack pointer to end of locals block */
	sp = &lcl[meth->localsz + meth->stacksz];

	restart:
	wide = 0;
	code = (bytecode*)meth->c.bcode.code;

	/* Finally we get to actually execute the machine */
	for (;;) {
		assert(npc < meth->c.bcode.codelen);
		pc = npc;
		mjbuf.pc = pc;
		npc = pc + insnLen[code[pc]];

		/* Handle WIDEned instructions */
		if (wide) {
			switch(code[pc]) {
			case ILOAD:
			case FLOAD:
			case ALOAD:
			case LLOAD:
			case DLOAD:
			case ISTORE:
			case FSTORE:
			case ASTORE:
			case LSTORE:
			case DSTORE:
				npc += 1;
				break;
			case IINC:
				npc += 2;
				break;
			default:
				fprintf(stderr, "Badly widened instruction");
				throwException(VerifyError);
				break;
			}
		}

		switch (code[pc]) {
		default:
			fprintf(stderr, "Unknown bytecode %d\n", code[pc]);
			throwException(VerifyError);
			break;
#include "kaffe.def"
		}
	}
	end:

	/* Unsync. if required */
	if (mobj != 0) {
		soft_monitorexit((Hjava_lang_Object* /*volatile*/)mobj);
	}       
	if (currentThread != NULL && unhand(currentThread)->PrivateInfo != 0) {
		TCTX(currentThread)->exceptPtr = mjbuf.prev;
	}

	DBG(DBG_METHOD_CALL,
	    dbgMethodCallDone();
	    if (callDepth < MAXDEPTHPRINT)
	    {
		    printf("method %s.%s%s:%s\n",
			   meth->class->name->data,
			   meth->name->data, meth->signature->data,
			   dbgMethodReturnVal(meth->signature, retval));
	    }
	    );
	
}

