/*
 * thread.c
 * Thread support.
 *
 * Copyright (c) 1996 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>, 1996.
 */
#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "config-io.h"
#include "config-signal.h"
#include "jtypes.h"
#include "access.h"
#include "object.h"
#include "constants.h"
#include "classMethod.h"
#include "baseClasses.h"
#include "lookup.h"
#include "thread.h"
#include "locks.h"
#include "exception.h"
#include "support.h"
#include "external.h"
#include "errors.h"
#include "lerrno.h"
#include "gc.h"
#include "md.h"
#include "debug.h"

Hjava_lang_Class* ThreadClass;
Hjava_lang_Class* ThreadGroupClass;
Hjava_lang_Thread* currentThread;
Hjava_lang_Thread* garbageman;
Hjava_lang_Thread* finalman;
struct Hjava_lang_ThreadGroup* standardGroup;
int threadStackSize;

static void firstStartThread(void);
static void createInitialThread(char*);
static Hjava_lang_Thread* createDaemon(void*, char*);

static void walkThread(void*, uint32);
static void finalizeThread(void*);
static gcFuncs gcThread = { walkThread, finalizeThread };

/****/
static int numberOfThreads;
static int talive;
static int tdaemon;
static void alarmException(int);
static void allocThreadCtx(Hjava_lang_Thread*, int);
/***/
void initThreads(void);
void startThread(Hjava_lang_Thread* );
static void createInitialThread(char*);
static Hjava_lang_Thread* createDaemon(void*, char*);


/*
 * Pick the right thread.c file based on what configuration we've
 * compiled for.  These files define initThreads(), startThread(),
 * createInitialThread(), and createDaemon().
 */
#if defined(HAVE_PACKAGE_TJWASSOC_CO_UK) && defined(HAVE_PACKAGE_SENSEI_CO_UK)
# error "Must select either Sun-style native libraries, *OR* Kore native libraries."
#elif defined(HAVE_PACKAGE_TJWASSOC_CO_UK)
# include "thread-sun.c"
#elif defined(HAVE_PACKAGE_SENSEI_CO_UK)
# include "thread-kore.c"
#else
# error "Neither package/tjwassoc.co.uk/ nor package/sensei.co.uk/ selected."
#endif

#ifdef DEBUG
char *
threadName(Hjava_lang_Thread* tid)
{
	static char cname[128];

	/*
	 * Tim's threads use HArrayOfChar for the name, whilst
	 * Kore threads use Hjava_lang_String for the name.
	 */
#if defined(HAVE_PACKAGE_TJWASSOC_CO_UK)
	int i;
	jchar *chrs = ARRAY_DATA(unhand(tid)->name);
	for (i=0; i<ARRAY_SIZE(unhand(tid)->name); i++)
	{
		cname[i] = chrs[i];
	}
	cname[i] = '\0';
#else
	if (!javaString2CString(unhand(tid)->name, cname, 128))
	{
		return "Unnamed";
	}
#endif
	if (cname[0] == '\0')
	{
		return "Unnamed";
	}

	return cname;
}
#endif


/*
 * Stop a thread from running and terminate it.
 */
void
stopThread(Hjava_lang_Thread* tid)
{
DBG(DBG_THREADOP,
	printf("stopThread %p:%s\n", tid, threadName(tid)); )	    

	if (currentThread == tid) {
		SignalError(0, "java.lang.ThreadDeath", "");
	}
	else {
		THREAD_STOP(tid);
	}
}

/*
 * All threads start here.
 */
static
void
firstStartThread(void)
{
	THREAD_INIT();

DBG(DBG_THREADOP,
    	printf("firstStartThread %p:%s\n", 
	       CURRENTTHREAD(), threadName(CURRENTTHREAD())); )

	/* Find the run()V method and call it */
	do_execute_java_method(0, (Hjava_lang_Object*)CURRENTTHREAD(), "run", "()V", 0, 0);
	do_execute_java_method(0, (Hjava_lang_Object*)CURRENTTHREAD(), "exit", "()V", 0, 0);

	THREAD_EXIT();
}

/*
 * Walk a thread, context and stack.
 */
static
void
walkThread(void* base, uint32 size)
{
	Hjava_lang_Thread* tid;

	tid = (Hjava_lang_Thread*)base;

#ifdef notyet
        markObject (.. class ..);
#endif

	markObject(unhand(tid)->name);
	markObject(unhand(tid)->threadQ);
	markObject(unhand(tid)->target);
	markObject(unhand(tid)->group);

	/* Since this thread my be extended by another class, walk any
	 * remaining data.
	 */
	gcStats.markedmem -= (size - sizeof(Hjava_lang_Thread));
	walkConservative(tid+1, size - sizeof(Hjava_lang_Thread));

	if (unhand(tid)->PrivateInfo != 0) {
		ctx* ct;
		ct = TCTX(tid);
		/* Nothing in context worth looking at except the stack */
#if defined(STACK_GROWS_UP)
		walkConservative(ct->stackBase, ct->restorePoint - ct->stackBase);
#else
		walkConservative(ct->restorePoint, ct->stackEnd - ct->restorePoint);
#endif
	}
}

/*
 * Finalize a thread.
 *  This is chiefly to free the thread context.
 */
static
void
finalizeThread(void* mem)
{
	Hjava_lang_Thread* tid;
        Method* final;

	tid = (Hjava_lang_Thread*)mem;

DBG(DBG_THREADOP,
	printf("finalizeThread %p:%s\n", tid, threadName(tid)); )

	THREAD_FREE(tid);

	/* Call thread finalizer if it has one (can it have one?) */
	final = findMethod(OBJECT_CLASS(&tid->base), final_name, void_signature);
	if (final != NULL) {
		CALL_KAFFE_METHOD(final, &tid->base);
	}
}

void
walkLiveThreads(void)
{
	Hjava_lang_Thread* tid;

	for (tid = liveThreads; tid != NULL; tid = TCTX(tid)->nextlive) {
		walkMemory(tid);
	}
}

/*
 * Yield process to another thread of equal priority.
 */
void
yieldThread()
{
DBG(DBG_THREADOP,
        printf("yieldThread %p:%s\n", CURRENTTHREAD(), threadName(CURRENTTHREAD())); )

	THREAD_YIELD();
}

/*
 * Put a thread to sleep.
 */
void
sleepThread(jlong time)
{
DBG(DBG_THREADOP,
	printf("sleepThread %p:%s\n", CURRENTTHREAD(), threadName(CURRENTTHREAD())); )

	THREAD_SLEEP(time);
}

/*
 * Is this thread alive?
 */
bool
aliveThread(Hjava_lang_Thread* tid)
{
	bool status;

DBG(DBG_THREADOP,
    	printf("aliveThread: %p\n", tid);				)

	THREAD_ALIVE(tid, status);

	return (status);
}

/*
 * How many stack frames have I invoked?
 */
jint
framesThread(Hjava_lang_Thread* tid)
{
	jint count;

	THREAD_FRAMES(tid, count);

	return (count);
}

/*
 * If we're not using native threads, include some extra thread support.
 */
#if !defined(USE_NATIVE_THREADS)
#include "thread-internal.c"
#endif
