/*
 * NativeIO.c
 *
 * Copyright (c) 1997 The University of Utah and
 * the Computer Systems Laboratory (CSL).  All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 *      Author: Patrick Tullmann <tullmann@cs.utah.edu>, University of Utah CSL
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "config-io.h"
#include "files.h"
#include "native.h"
#include "defs.h"
#include "java.io.stubs/FileDescriptor.h"
#include "java.io.stubs/NativeIO.h"
#include "kthread.h"
#include "../../../../kaffe/kaffevm/debug.h"

/*
 * Returns number of bytes that can be read without blocking.
 */
jint java_io_NativeIO_available(struct Hjava_io_FileDescriptor* fd)
{
	/* Copied from packages/tjwassoc.co.uk/APIcore/lib/java.io/FileInputStream.c */
	int nr;
	int r;

#if defined(HAVE_IOCTL) && defined(FIONREAD)
	r = ioctl(unhand(fd)->fd, FIONREAD, &nr);
	if (r < 0) {
		nr = 0;
	}
#else
	/* This uses select() to work out if we can read - but what
	 * happens at the end of file?
	 */
	static struct timeval tm = { 0, 0 };
	int fd;
	fd_set rd;

	fd = unhand(fd)->fd;
	FD_ZERO(&rd);
	FD_SET(fd, &rd);
	r = select(fd+1, &rd, NULL, NULL, &tm);
	if (r == 1) {
		nr = 1;
	}
	else {
		nr = 0;
	}
#endif

	return (nr);
}

/*
 * Read into bytes
 */
jint 
java_io_NativeIO_read(struct Hjava_io_FileDescriptor* fd,
		      HArrayOfByte* bytes, jint offset, jint length)
{
        jint ret;

	DBG(DBG_NATIVE_IO,
	    printf("NativeIO.read(fd=%d, bytes=%p, offset=%d, length=%d);\n",
		   unhand(fd)->fd, bytes, offset, length)
	    );

        ret = read(unhand(fd)->fd, &unhand(bytes)->body[offset], length);
	DBG(DBG_NATIVE_IO,
	    printf("NativeIO.read()=%d\n", ret)
	    );
        if (ret < 0) {
		DBG(DBG_NATIVE_IO,
		    printf("NativeIO.read() -> %s\n", SYS_ERROR)
		    );
                SignalError(NULL, "java.io.IOException", SYS_ERROR);
        }
        return (ret > 0 ? ret : -1); /* Can't ever return 0 */
}

jint 
java_io_NativeIO_write(struct Hjava_io_FileDescriptor* fd,
		       HArrayOfByte* bytes, jint offset, jint length)
{
	jint ret;

	DBG(DBG_NATIVE_IO,
	    printf("NativeIO.write(fd=%d, bytes=%p, offset=%d, length=%d);\n",
		   unhand(fd)->fd, bytes, offset, length)
	    );
	ret = write(unhand(fd)->fd, &unhand(bytes)->body[offset], length);
	DBG(DBG_NATIVE_IO,
	    printf("NativeIO.write()=%d\n", ret)
	    );
	if (ret < 0) {
		SignalError(NULL, "java.io.IOException", SYS_ERROR);
	}
	return ret; /* XXX what to return here? */
}

jint 
java_io_NativeIO_skip(struct Hjava_io_FileDescriptor* fd,
		      jint off)
{
#define SKIPBUFSZ 128
	off_t orig;
	off_t ret;
	char buffer[SKIPBUFSZ];
	int count;

	DBG(DBG_NATIVE_IO,
	    printf("NativeIO.skip(fd=%d, off=%d)\n",
		   (int)(unhand(fd)->fd), off);
	    );
	
	orig = lseek(unhand(fd)->fd, (off_t)0, SEEK_CUR);
	ret = lseek(unhand(fd)->fd, jlong2off_t((jlong)off), SEEK_CUR);
	if (ret >= 0) {
		return (jint)(off_t2jlong(ret-orig));
	}

	/* Not seekable - try just reading. */
	ret = 0;
	while (off > 0) {
		count = read(unhand(fd)->fd, buffer, (off%SKIPBUFSZ));
		if (count == 0) {	/* Reached EOF. */
			break;
		}
		if (count < 0) {
			SignalError(NULL, "java.io.IOException", SYS_ERROR);
		}
		ret += count;
		off -= count;
	}
	return (jint)(off_t2jlong(ret));
#undef SKIPBUFSZ
}

void 
java_io_NativeIO_close(struct Hjava_io_FileDescriptor* fd)
{
	int ret;

	if (unhand(fd)->fd >= 0) {
		DBG(DBG_NATIVE_IO,
		    printf("NativeIO.close(fd=%d)\n", (int)(unhand(fd)->fd))
		    );
		ret = close(unhand(fd)->fd);
		unhand(fd)->fd = -1;
		if (ret < 0) {
			SignalError(NULL, "java.io.IOException", SYS_ERROR);
		}
	}
}

void
java_io_NativeIO_open_read(struct Hjava_io_FileDescriptor* fd,
			   struct Hjava_lang_String* name)
{
	char strbuf[MAXPATHLEN];

	assert(fd);

	javaString2CString(name, strbuf, MAXPATHLEN);
	DBG(DBG_NATIVE_IO,
	    printf("NativeIO.open_read(%s)\n", strbuf)
	    );
	unhand(fd)->fd = (jint) open(strbuf, O_RDONLY|O_BINARY, 0);
	DBG(DBG_NATIVE_IO, 
	    printf("NativeIO.open_read(%s)->fd=%d\n", strbuf, (int)unhand(fd)->fd)
	    );
	if (unhand(fd)->fd < 0) {
		DBG(DBG_NATIVE_IO,
		    printf("NativeIO.open_read(%s)->err=%s\n", strbuf, SYS_ERROR)
		    );
		SignalError(0, "java.io.IOException", SYS_ERROR);
	}
	
}

void
java_io_NativeIO_open_write(struct Hjava_io_FileDescriptor* fd,
			    struct Hjava_lang_String* name)
{
	char strbuf[MAXPATHLEN];

	assert(fd);

	javaString2CString(name, strbuf, MAXPATHLEN);
	DBG(DBG_NATIVE_IO,
	    printf("NativeIO.open_write(%s)\n", strbuf)
	    );
	unhand(fd)->fd = (jint) open(strbuf, O_WRONLY|O_CREAT|O_BINARY|O_TRUNC, 0666);
	DBG(DBG_NATIVE_IO,
	    printf("NativeIO.open_write(%s)->fd=%d\n", strbuf, (int)unhand(fd)->fd)
	    );
	if (unhand(fd)->fd < 0) {
		DBG(DBG_NATIVE_IO,
		    printf("NativeIO.open_write(%s)->err=%s\n", strbuf, SYS_ERROR)
		    );
		SignalError(0, "java.io.IOException", SYS_ERROR);
	}
	
}

void
java_io_NativeIO_open_read_write(struct Hjava_io_FileDescriptor* fd,
				 struct Hjava_lang_String* name)
{
	unimp("Kore function not implemented");
}

jint 
java_io_NativeIO_ftell(struct Hjava_io_FileDescriptor* fd)
{
	off_t offset;

	offset = lseek(unhand(fd)->fd, 0, SEEK_CUR);
	if (offset < 0) {
		SignalError(NULL, "java.io.IOException", SYS_ERROR);
	}
	return (off_t2jlong(offset));
}

jint
java_io_NativeIO_fseek(struct Hjava_io_FileDescriptor* fd,
		       jint wantedOffset, jint whence)
{
	int actualOffset;
	int seekdir = -1;

	/*
	 * XXX shouldn't wantedOffset and return value both be longs?
	 */

	switch (whence) 
	{
	case java_io_NativeIO_SEEK_SET:
		seekdir = SEEK_SET;
		break;
	case java_io_NativeIO_SEEK_CUR:
		seekdir = SEEK_CUR;
		break;
	case java_io_NativeIO_SEEK_END:
		seekdir = SEEK_END;
		break;
	}

	DBG(DBG_NATIVE_IO,
	    printf("NativeIO.fseek(fd=%d, wantedOffset=%d, whence=%s)\n",
		   unhand(fd)->fd, wantedOffset,
		   (whence==SEEK_SET)?"set":
		   ((whence==SEEK_CUR)?"current":
		    ((whence==SEEK_END)?"end":"wha?")))
	    );

	actualOffset = lseek(unhand(fd)->fd, (off_t)wantedOffset, seekdir);
	if (actualOffset < 0) {
		SignalError(NULL, "java.io.IOException", SYS_ERROR);
	}

	return (jint)actualOffset;
}

jbool
java_io_NativeIO_canRead(struct Hjava_lang_String* path)
{
	char str[MAXPATHLEN];
	int r;

	javaString2CString(path, str, sizeof(str));
	r = access(str, R_OK);
	return (r < 0 ? 0 : 1);
}

jbool
java_io_NativeIO_canWrite(struct Hjava_lang_String* path)
{
	char str[MAXPATHLEN];
	int r;

	javaString2CString(path, str, sizeof(str));
	r = access(str, W_OK);
	return (r < 0 ? 0 : 1);
}

jbool
java_io_NativeIO_exists(struct Hjava_lang_String* path)
{
	char str[MAXPATHLEN];
	struct stat statBuf;

	javaString2CString(path, str, sizeof(str));

	/* File exists if stat doesn't err on this path */
	return (stat(str, &statBuf) < 0) ? 0 : 1;
}

jbool
java_io_NativeIO_isDirectory(struct Hjava_lang_String* path)
{
	unimp("Kore function not implemented");
	return 0;
}

jbool
java_io_NativeIO_isFile(struct Hjava_lang_String* path)
{
	struct stat buf;
	char str[MAXPATHLEN];
	int r;

	javaString2CString(path, str, MAXPATHLEN);

	r = stat(str, &buf);
	if (r == 0 && S_ISREG(buf.st_mode)) {
		return 1;
	} else {
		return 0;
	}
}

jlong
java_io_NativeIO_lastModified(struct Hjava_lang_String* path)
{
	unimp("Kore function not implemented");
	return 0;
}

jint
java_io_NativeIO_fileSize(struct Hjava_lang_String* path)
{
	unimp("Kore function not implemented");
	return 0;
}

jbool
java_io_NativeIO_delete(struct Hjava_lang_String* path)
{
	unimp("Kore function not implemented");
	return 0;
}

jbool
java_io_NativeIO_rename(struct Hjava_lang_String* from,
			struct Hjava_lang_String* to)
{
	unimp("Kore function not implemented");
	return 0;
}

jbool
java_io_NativeIO_mkdir(struct Hjava_lang_String* path)
{
	unimp("Kore function not implemented");
	return 0;
}

HArrayOfObject* /* String[], actually */
java_io_NativeIO_dirList(struct Hjava_lang_String* path)
{
	unimp("Kore function not implemented");
	return 0;
}

